# kernel selection spoke classes
#
# Copyright (C) 2021 OpenAnolis Community
#
import re
import gi
import time

gi.require_version("Gtk", "3.0")
gi.require_version("Pango", "1.0")

from gi.repository import Gtk, Pango

from pyanaconda.flags import flags
from pyanaconda.core.i18n import _, C_, CN_
from pyanaconda.threading import threadMgr, AnacondaThread
from pyanaconda.payload.manager import payloadMgr, PayloadState
from pyanaconda.payload.base import Payload
from pyanaconda.payload.errors import NoSuchGroup, PayloadError, DependencyError

from pyanaconda.core import util, constants

from pyanaconda.ui.communication import hubQ
from pyanaconda.ui.gui.spokes import NormalSpoke
from pyanaconda.ui.gui.spokes.lib.detailederror import DetailedErrorDialog
from pyanaconda.ui.gui.utils import blockedHandler, escape_markup
from pyanaconda.core.async_utils import async_action_wait
from pyanaconda.ui.categories.software import SoftwareCategory

from pyanaconda.anaconda_loggers import get_module_logger
log = get_module_logger(__name__)

import sys, copy

__all__ = ["KernelSelectionSpoke"]

class KernelSelectionSpoke(NormalSpoke):
    builderObjects = ["kernelWindow"]
    mainWidgetName = "kernelWindow"
    uiFile = "spokes/kernel_selection.glade"
    help_id = "KernelSelectionSpoke"

    category = SoftwareCategory

    #icon = "applications-system"
    icon = "package-x-generic-symbolic"
    title = CN_("GUI|Spoke", "_Kernel Selection")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._error_msgs = None
        self._error = False

        self._orig_kernel = None

        self._kickstarted = flags.automatedInstall and self.data.packages.seen

        self._kernelListBox = self.builder.get_object("kernelListBox")
        self._kernelViewport = self.builder.get_object("kernelViewport")
        self._kernelListBox.set_focus_vadjustment(Gtk.Scrollable.get_vadjustment(self._kernelViewport))

        self._fakeRadio = Gtk.RadioButton(group=None)
        self._fakeRadio.set_active(True)

        # Register event listeners to update our status on payload events
        payloadMgr.add_listener(PayloadState.ERROR, self._payload_error)

        payloadMgr.add_listener(PayloadState.DOWNLOADING_PKG_METADATA,
                                self._downloading_package_md)
        payloadMgr.add_listener(PayloadState.DOWNLOADING_GROUP_METADATA,
                                self._downloading_group_md)

    # Payload event handlers
    def _downloading_package_md(self):
        # Reset the error state from previous payloads
        self._error = False
        hubQ.send_message(self.__class__.__name__, _(constants.PAYLOAD_STATUS_PACKAGE_MD))

    def _downloading_group_md(self):
        hubQ.send_message(self.__class__.__name__, _(constants.PAYLOAD_STATUS_GROUP_MD))

    def _payload_error(self):
        self._error = True
        hubQ.send_message(self.__class__.__name__, payloadMgr.error)

    @property
    def available_kernels(self):
        return self.payload.available_kernels

    @property
    def current_kernel(self):
        return self.payload.current_kernel

    @current_kernel.setter
    def current_kernel(self, value):
        self.payload.current_kernel = value

    def apply(self):
        self._apply()

    def _apply(self):
        hubQ.send_not_ready(self.__class__.__name__)
        hubQ.send_not_ready("SourceSpoke")
        threadMgr.add(AnacondaThread(name="AnaCheckKernel", target=self.checkKernelSelection))

    def checkKernelSelection(self):
        if self._kickstarted:
            kernel_pattern=re.compile(r"kernel-[4,5].\d+.\d+")
            for package in self.payload.data.packages.packageList:
                if kernel_pattern.search(package):
                    self.current_kernel = package.split('-')[1]

        # we do this only kernel changed
        if self.changed:
            kernel_keys = self.available_kernels.keys()
            for kernel in kernel_keys:
                kernel_sub_keys = list(self.available_kernels[kernel].keys())
                if kernel == self.current_kernel:
                    include_kernel = self.available_kernels[kernel][kernel_sub_keys[-1]]
                    self.payload.data.packages.packageList.extend(include_kernel)
                    for package in include_kernel:
                        if package in self.payload.data.packages.excludedList:
                            self.payload.data.packages.excludedList.remove(package)
                else:
                    exclude_kernel = self.available_kernels[kernel][kernel_sub_keys[-1]]
                    self.payload.data.packages.excludedList.extend(exclude_kernel)
                    for package in exclude_kernel:
                        if package in self.payload.data.packages.packageList:
                            self.payload.data.packages.packageList.remove(package)

            # store kernel selection
            self._orig_kernel = self.current_kernel

        hubQ.send_ready(self.__class__.__name__, False)
        hubQ.send_ready("SourceSpoke", False)

    def initialize(self):
        super().initialize()
        self.initialize_start()
        threadMgr.add(AnacondaThread(name="AnaKernelWatcher", target=self._initialize))

    def _initialize(self):
        threadMgr.wait(constants.THREAD_PAYLOAD)

        if not self._kickstarted:
            self._first_refresh()

        hubQ.send_ready(self.__class__.__name__, False)

        if not self._error:
            self._apply()

        self.initialize_done()

    @property
    def completed(self):
        processingDone = bool(not threadMgr.get("AnaCheckKernel") and
                              not threadMgr.get(constants.THREAD_PAYLOAD) and
                              self.current_kernel is not None)

        if processingDone:
            return True

    @property
    def changed(self):
        if self.current_kernel == self._orig_kernel:
            return False
        else:
            return True

    @property
    def mandatory(self):
        return True

    @property
    def ready(self):
        return bool(not threadMgr.get("AnaKernelWatcher") and
                    not threadMgr.get(constants.THREAD_PAYLOAD) and
                    not threadMgr.get("AnaCheckKernel") and 
                    self.payload.base_repo is not None)

    @property
    def showable(self):
        return isinstance(self.payload, Payload)

    @property
    def status(self):
        if not self.ready:
            return _("Installation source not set up")

        if self._error_msgs:
            return _(self._error_msgs)

        if not flags.automatedInstall:
            if not self.current_kernel:
                return _("Please confirm kernel selection")

        return self.current_kernel


    def _add_row(self, listbox, name, desc, button, clicked):
        row = Gtk.ListBoxRow()
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        button.set_valign(Gtk.Align.START)
        button.connect("toggled", clicked, row)
        box.add(button)

        label = Gtk.Label(label="<b>%s</b>\n%s" % (escape_markup(name), escape_markup(desc)),
                          use_markup=True, wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR,
                          hexpand=True, xalign=0, yalign=0.5)
        box.add(label)

        row.add(box)
        listbox.insert(row, -1)

    @async_action_wait
    def _first_refresh(self):
        self.refresh()
        return True

    def refresh(self):
        super().refresh()

        threadMgr.wait(constants.THREAD_PAYLOAD)

        self._error_msgs = self.payload.detectMultiKernel()
        if not self._error_msgs:
            # use latest kernel as default, normally it's anck
            self.current_kernel = list(self.available_kernels.keys())[-1]

        self._clear_listbox(self._kernelListBox)

        kernel_keys = self.available_kernels.keys()
        for kernel in kernel_keys:
            kernel_sub_keys = list(self.available_kernels[kernel].keys())
            radio = Gtk.RadioButton(group=self._fakeRadio)
            if not self.current_kernel:
                self.current_kernel = kernel
            radio.set_active(self.current_kernel and kernel == self.current_kernel)
            if kernel == '4.18.0':
                kernel_desc = _("RHCK")
                kernel_info = _("Compatible with RHEL")
            elif kernel == '4.19.91':
                kernel_desc = _("ANCK")
                kernel_info = _("Support Anolis OS verified platform")
            else:
                kernel_desc = ''
                kernel_info = ''

            self._add_row(self._kernelListBox, "%s (%s)" % (kernel, kernel_desc), "%s (%s)" % (kernel_info, self.available_kernels[kernel][kernel_sub_keys[-1]][0]), radio, self.on_radio_button_toggled)

        self._kernelListBox.show_all()

    def _clear_listbox(self, listbox):
        for child in listbox.get_children():
            listbox.remove(child)
            del(child)

    # Signal handlers
    def on_radio_button_toggled(self, radio, row):
        # If the radio button toggled to inactive, don't reactivate the row
        if not radio.get_active():
            return
        row.activate()

    def on_kernel_activated(self, listbox, row):
        box = row.get_children()[0]
        button = box.get_children()[0]

        with blockedHandler(button, self.on_radio_button_toggled):
            button.set_active(True)

        kernel_keys = list(self.available_kernels.keys())
        self.current_kernel=kernel_keys[row.get_index()]
